// Copyright (c) 2004-2005 Lutz Latta
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#define GLH_NVEB_USING_CGGL
#define GLH_EXT_SINGLE_FILE
#define _USE_MATH_DEFINES

#include <math.h>

#if defined(WIN32)
#  include <windows.h>
#endif

#include <Cg/cgGL.h>
#include <glh/glh_extensions.h>
#include <glh/glh_glut.h>
#include <glh/glh_linear.h>
#include "shared/nv_png.h"
#include "shared/MovieMaker.h"

#include "ParticleSystem.h"

using namespace glh;

ParticleSystem* particleSystem;

// view parameters
float2 previousMousePosition;
int buttonState = 0;
float3 viewerPosition(0.0f, 0.0f, 0.0f);
float2 viewerRotation(0.0f, 0.0f);
float2 viewerRotationSpeed(0.0f, 3.0f);

enum ControlMode { M_VIEW, M_MOVE_SPHERE, M_MOVE_EMITTER, M_ROTATE_EMITTER, M_MOVE_LOCAL_FORCE };
ControlMode mode = M_VIEW;
int moveEmitterIndex = 0;
int moveLocalForceIndex = 0;

float lastTime;

MovieMaker movieMaker;
bool isMovieMaking = false;


bool toggle[256];
float3 spherePosition;
vector<float3> emitterPositions;
vector<float3> emitterLookatPoints;
vector<float3> localForcePositions;

// check for OpenGL errors
void checkGLErrors(const char *s)
{
	GLenum error;
	while ((error = glGetError()) != GL_NO_ERROR) {
		fprintf(stderr, "%s: error - %s\n", s, (char *) gluErrorString(error));
	}
}

void initParticleSystemVariables()
{
	viewerPosition = particleSystem->GetViewerPosition();
	viewerRotation = particleSystem->GetViewerRotation();
	spherePosition = particleSystem->GetSpherePosition();

	emitterPositions.resize(particleSystem->GetNumOfEmitters());
	emitterLookatPoints.resize(particleSystem->GetNumOfEmitters());

	int i;
	for (i = 0; i < particleSystem->GetNumOfEmitters(); i++)
	{
		emitterPositions[i] = particleSystem->GetEmitterPosition(i);
		emitterLookatPoints[i] = emitterPositions[i] + particleSystem->GetEmitterDirection(i);
	}
	moveEmitterIndex = 0;

	localForcePositions.resize(particleSystem->GetNumOfLocalForces());
		for (i = 0; i < particleSystem->GetNumOfLocalForces(); i++)
			localForcePositions[i] = particleSystem->GetLocalForcePosition(i);
	moveLocalForceIndex = 0;

	const bool* menuKeys = particleSystem->GetMenuKeys();
	if (!menuKeys[0]) // If menu key zero is on, this means no keys were loaded from config.
	{
		for (char c = 'a'; c <= 'z'; c++)
			toggle[c] = menuKeys[c];
		//memcpy(toggle, menuKeys, sizeof(toggle);
	}
}

// Draw a grid of lines
void drawWireGrid(int rows, int cols)
{
	int i;
	glLineWidth(1.0);

	glBegin(GL_LINES);
	for(i=0; i<=rows; i++) {
		glVertex2f(0.0, (float) i / rows);
		glVertex2f(1.0, (float) i / rows);
	}
	for(i=0; i<=cols; i++) {
		glVertex2f((float) i / cols, 0.0);
		glVertex2f((float) i / cols, 1.0);
	}
	glEnd();
}

void drawText(const char* text, int x, int y, bool drawTopDown)
{
	int width = glutGet(GLUT_WINDOW_WIDTH);
	int height = glutGet(GLUT_WINDOW_HEIGHT);

	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, width - 1, 0, height - 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(1, 1, 1);

	if (drawTopDown)
		y = height - y;
	glRasterPos2i(x, y);

	for (int i = 0; text[i]; i++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, text[i]);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glEnable(GL_DEPTH_TEST);
}

void display(void)
{	
	float time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;

	static float lastSecondTime = time;
	static int lastFrameNum = 0;

	static int frameNum = 0;
	frameNum++;

	static float lastFPS = 0.0f;
	if (time > lastSecondTime + 1.0)
	{
		lastFPS = (frameNum - lastFrameNum) / (time - lastSecondTime);

		lastSecondTime = time;
		lastFrameNum = frameNum;
	}

	if (toggle['a'])
	{
		// Simulate particle system
#if 0
		particleSystem->TimeStep(time - lastTime);
#else
		particleSystem->TimeStep(1.0f);
#endif
	}

	lastTime = time;

	float4 c = particleSystem->GetClearColor();
	glClearColor(c[0], c[1], c[2], c[3]);
	//glClearColor(0.2, 0.5, 0.9, 1.0);
	//glClearColor(100.0/255, 130.0/255, 150.0/255, 1.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// view transform
	if (toggle['u'])
	{
		// Animate camera.
		viewerRotation += viewerRotationSpeed;
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(viewerPosition[0], viewerPosition[1], viewerPosition[2]);
	glRotatef(viewerRotation[0], 1.0, 0.0, 0.0);
	glRotatef(viewerRotation[1], 0.0, 1.0, 0.0);

	// Calculate viewer position.
	float4x4 mvMat;
	glGetFloatv(GL_MODELVIEW_MATRIX, mvMat.m);
	mvMat = mvMat.inverse();
	float3 vp(mvMat.m[12], mvMat.m[13], mvMat.m[14]);
	particleSystem->SetViewerPosition(-vp);
	particleSystem->SetViewerRotation(viewerRotation);

	// grid
	if (toggle['g'])
	{
		glColor3f(0.3, 0.3, 0.3);
		glPushMatrix();
		glRotatef(90.0, 1.0, 0.0, 0.0);
		glScalef(8.0, 8.0, 8.0);
		glTranslatef(-0.5, -0.5, 0.0);
		drawWireGrid(20, 20);
		glScalef(1 / 8.0, 1 / 8.0, 1 / 256.0);
		glPopMatrix();
	}
	if (toggle['h'])
		particleSystem->DisplayTerrain();
	if (toggle['l'])
		particleSystem->DisplayFlowField();

	// Draw helper objects.
	if (toggle['s'])
	{
		// Draw collision sphere.
		particleSystem->DisplayCollisionSphere();

		// Draw local forces.
		for (int i = 0; i < particleSystem->GetNumOfLocalForces(); i++)
		{
			glPushMatrix();
			glTranslatef(localForcePositions[i][0], localForcePositions[i][1], localForcePositions[i][2]);
			if (mode == M_MOVE_LOCAL_FORCE && i == moveLocalForceIndex)
				glColor3f(1.0, 1.0, 0.0);
			else
				glColor3f(0.1, 0.3, 0.0);
			glutSolidSphere(0.05, 10, 10);
			glPopMatrix();
		}

		// Draw look-at point of particle emitter.
		if (mode == M_ROTATE_EMITTER)
		{
			glPushMatrix();
			glTranslatef(emitterLookatPoints[moveEmitterIndex][0],
				emitterLookatPoints[moveEmitterIndex][1], emitterLookatPoints[moveEmitterIndex][2]);
			glColor3f(0.9, 0.9, 0.9);
			glutSolidSphere(0.02, 3, 3);
			glPopMatrix();
		}
	}

	// draw particle system
	//particleSystem->ReadBack();
	//particleSystem->Display();
	if (toggle['D'])
		particleSystem->DisplayPoints();

	if (toggle['t'])
		particleSystem->DisplayTextures();

	checkGLErrors("display");

	// Display info messages
	if (!isMovieMaking)
	{
		char buf[256];
		sprintf(buf, "Particle system: %dx%d, currently active: %d", particleSystem->GetWidth(), particleSystem->GetHeight(), particleSystem->GetNumOfActiveParticles());
		drawText(buf, 20, 30, true);
		sprintf(buf, "%05.2f fps", lastFPS);
		drawText(buf, 20, 50, true);

		drawText("Press key 1, 2, 3, 4 or 0 to load example setups. Right click for more options.", 20, 20, false);
	}

	glutSwapBuffers();

	if (isMovieMaking)
	{
#ifdef _WIN32
		Sleep(50);
#endif
		movieMaker.Snap();
	}
}

void motion(int x, int y)
{
	float2 mousePosition(x, y);
	float2 delta(mousePosition - previousMousePosition);
	previousMousePosition = mousePosition;

	float3 dir;
	switch (mode)
	{
	case M_VIEW:
		if (buttonState == 3)
		{
			// left+middle = zoom
			viewerPosition[2] -= delta[1] / 100.0f;
		} 
		else if (buttonState == 2)
		{
			// middle = translate
			viewerPosition += float3(delta[0], -delta[1], 0.0f) / 100.0f;
		}
		else if (buttonState == 1)
		{
			// left = rotate
			viewerRotation += float2(delta[1], delta[0]) / 5.0f;
		}
		break;

	case M_MOVE_SPHERE:
		if (buttonState == 1)
		{
			spherePosition += float3(delta[0], -delta[1], 0.0f) / 100.0f;
		}
		else if (buttonState == 2)
		{
			spherePosition += float3(delta[0], 0.0f, delta[1]) / 100.0f;
		}
		particleSystem->SetSpherePosition(spherePosition);
		break;

	case M_MOVE_EMITTER:
		if (buttonState == 1)
		{
			emitterPositions[moveEmitterIndex] += float3(delta[0], -delta[1], 0.0f) / 100.0f;
			emitterLookatPoints[moveEmitterIndex] += float3(delta[0], -delta[1], 0.0f) / 100.0f;
		}
		else if (buttonState == 2)
		{
			emitterPositions[moveEmitterIndex] += float3(delta[0], 0.0f, delta[1]) / 100.0f;
			emitterLookatPoints[moveEmitterIndex] += float3(delta[0], 0.0f, delta[1]) / 100.0f;
		}
		particleSystem->SetEmitterPosition(moveEmitterIndex, emitterPositions[moveEmitterIndex]);
		break;

	case M_ROTATE_EMITTER:
		if (buttonState == 1)
		{
			emitterLookatPoints[moveEmitterIndex] += float3(delta[0], -delta[1], 0.0f) / 100.0f;
		}
		else if (buttonState == 2)
		{
			emitterLookatPoints[moveEmitterIndex] += float3(delta[0], 0.0f, delta[1]) / 100.0f;
		}

		dir = emitterLookatPoints[moveEmitterIndex] - emitterPositions[moveEmitterIndex];
		dir.normalize();

		particleSystem->SetEmitterDirection(moveEmitterIndex, dir);
		break;

	case M_MOVE_LOCAL_FORCE:
		if (buttonState == 1)
		{
			localForcePositions[moveLocalForceIndex] += float3(delta[0], -delta[1], 0.0f) / 100.0f;
		}
		else if (buttonState == 2)
		{
			localForcePositions[moveLocalForceIndex] += float3(delta[0], 0.0f, delta[1]) / 100.0f;
		}
		particleSystem->SetLocalForcePosition(moveLocalForceIndex, localForcePositions[moveLocalForceIndex]);
		break;
	}

	glutPostRedisplay();
}


void mouse(int button, int state, int x, int y)
{
	int mods;

	if (state == GLUT_DOWN)
		buttonState |= 1 << button;
	else if (state == GLUT_UP)
		buttonState = 0;

	mods = glutGetModifiers();
	if (mods & GLUT_ACTIVE_SHIFT)
		buttonState = 2;
	else if (mods & GLUT_ACTIVE_CTRL)
		buttonState = 3;

	previousMousePosition = float2(x, y);
	glutPostRedisplay();
}


void key(unsigned char key, int x, int y)
{
	toggle[key] ^= 1;

	switch (key) {
	case '\033':
	case 'q':
		exit(0);
		break;
	case 'w':
		glPolygonMode(GL_FRONT_AND_BACK, toggle['w'] ? GL_LINE : GL_FILL);
		break;

	case 'v':
		mode = M_VIEW;
		break;

	case 'm':
		mode = M_MOVE_SPHERE;
		break;

	case 'e':
		mode = M_MOVE_EMITTER;
		break;

	case 'o':
		mode = M_ROTATE_EMITTER;
		break;

	case 'y':
		moveEmitterIndex = (moveEmitterIndex + 1) % particleSystem->GetNumOfEmitters();
		break;

	case 'f':
		mode = M_MOVE_LOCAL_FORCE;
		break;

	case 'd':
		moveLocalForceIndex = (moveLocalForceIndex + 1) % particleSystem->GetNumOfLocalForces();
		break;

	case 'r':
		particleSystem->Reset();
		break;

	case ' ':
		particleSystem->TimeStep(1.0f);
		break;

	case 'i':
		if (!isMovieMaking)
			movieMaker.StartCapture("particle.avi", 30);
		else
			movieMaker.EndCapture();
		isMovieMaking = !isMovieMaking;
		break;

	case '-':
		viewerRotationSpeed[0] -= 0.25f;
		printf("viewerRotationSpeed.x: %f\n", viewerRotationSpeed[0]);
		break;
	case '=':
		viewerRotationSpeed[0] += 0.25f;
		printf("viewerRotationSpeed.x: %f\n", viewerRotationSpeed[0]);
		break;
	case '_':
		viewerRotationSpeed[1] -= 0.25f;
		printf("viewerRotationSpeed.y: %f\n", viewerRotationSpeed[1]);
		break;
	case '+':
		viewerRotationSpeed[1] += 0.25f;
		printf("viewerRotationSpeed.y: %f\n", viewerRotationSpeed[1]);
		break;

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
		particleSystem->UnloadConfigSetup();
		particleSystem->LoadConfigSetup(key == '0' ? string() : string(1, key));
		particleSystem->Reset();

		initParticleSystemVariables();
		break;
	}

	glutPostRedisplay();
}

void specialKey(int key, int x, int y)
{
	switch(key)
	{
	case GLUT_KEY_F10:
	case GLUT_KEY_F1:
	case GLUT_KEY_F2:
	case GLUT_KEY_F3:
	case GLUT_KEY_F4:
		particleSystem->UnloadConfigSetup(false);
		particleSystem->LoadConfigSetup(key == GLUT_KEY_F10 ? string() : string(1, '1' + char(key - GLUT_KEY_F1)), false);

		initParticleSystemVariables();
		break;
	}

	glutPostRedisplay();
}

void idle(void)
{
	//if (toggle['a'])
		glutPostRedisplay();
}

void reshape(int w, int h)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (float) w / (float) h, 0.1, 100.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glViewport(0, 0, w, h);

	particleSystem->SetPerspective(60.0 * M_PI / 180.0, h);
}

void mainMenu(int i)
{
	key((unsigned char) i, 0, 0);
}

void initMenus(void)
{
	glutCreateMenu(mainMenu);
	glutAddMenuEntry("Load preset [0,1,2,3,4] (without resizing system [F10,F1,F2,F3,F4])", '0');
	glutAddMenuEntry("Toggle animation [a]", 'a');
	glutAddMenuEntry("Step animation [ ]", ' ');
	glutAddMenuEntry("Move/rotate viewer mode [v]", 'v');
	glutAddMenuEntry("Move sphere mode [m]", 'm');
	glutAddMenuEntry("Move emitter mode [e]", 'e');
	glutAddMenuEntry("Rotate emitter mode [o]", 'o');
	glutAddMenuEntry("Change emitter to move/rotate [y]", 'y');
	glutAddMenuEntry("Move local force mode [f]", 'f');
	glutAddMenuEntry("Change local force to move [d]", 'd');
	glutAddMenuEntry("Toggle display of data textures [t]", 't');
	glutAddMenuEntry("Toggle display of helper objects [s]", 's');
	glutAddMenuEntry("Toggle display of grid [g]", 'g');
	glutAddMenuEntry("Toggle display of height-field [h]", 'h');
	glutAddMenuEntry("Toggle display of flow-field [l]", 'l');
	glutAddMenuEntry("Toggle display of wireframe [w]", 'w');
	glutAddMenuEntry("Start/stop recording of video (\"particle.avi\") [i]", 'i');
	glutAddMenuEntry("Start/stop animation of camera [u]", 'u');
	glutAddMenuEntry("Change animation speed of camera [-, _, =, +]", '-');
	glutAddMenuEntry("Reset [r]", 'r');
	glutAddMenuEntry("Quit (esc)", '\033');
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

// get extension pointers
void getExts(void)
{
	if (!glh_init_extensions("GL_VERSION_1_2 "
		"GL_ARB_multitexture "
		"GL_ARB_vertex_buffer_object "
		"GL_NV_vertex_program "
		"GL_NV_fragment_program "
		"GL_NV_float_buffer "
		"GL_NV_vertex_array_range "
		"GL_NV_pixel_data_range "
		"GL_NV_primitive_restart "
		"GL_ARB_vertex_program " // for glVertexAttrib
 		"GL_EXT_pixel_buffer_object "
		)) {
			fprintf(stderr, "Error - required extensions were not supported: %s", glh_get_unsupported_extensions());
			exit(-1);
		}
}

void initGL()
{  
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_DITHER);

	checkGLErrors("initGL");
}

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(800, 600);
#if 0
    glutGameModeString("1280x720:32@60");
    glutEnterGameMode();
#else
	glutCreateWindow("MegaParticles");
#endif

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutKeyboardFunc(key);
	glutSpecialFunc(specialKey);
	glutIdleFunc(idle);

	toggle['g'] = true;
	toggle['s'] = true;
	toggle['a'] = true;
	toggle['t'] = false;
	toggle['h'] = false;
	toggle['l'] = false;
	toggle['u'] = false;

	// Unexposed key: Prevent display of particles
	toggle['D'] = true;

	initMenus();
	getExts();
	initGL();

	// create particle system
	particleSystem = new ParticleSystem();

	particleSystem->Reset();

	initParticleSystemVariables();

	lastTime = glutGet(GLUT_ELAPSED_TIME) / 1000.0;

	glutMainLoop();
	return 0;
}
