// Copyright (c) 2004-2005 Lutz Latta
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#define NOMINMAX
#define _USE_MATH_DEFINES

#include <string>
#include <sstream>
#include <limits>
#include <iostream>
using namespace std;

#include <math.h>
#include <Cg/cgGL.h>
#include <GL/glut.h>
#include <glh/glh_linear.h>
#include <glh/glh_extensions.h>
#include "shared/nv_png.h"
#include "shared/nv_dds.h"
#include "shared/array_texture.h"
#include "RenderToVA.h"
#include "RenderVertexArray.h"
#include "ParticleSystemConfig.h"

#include "ParticleSystem.h"

// Define this to 0 to use euler integration, otherwise verlet integration.
#define VERLET_INTEGRATION 0

// Define this to 1 to perform partial integration steps at the initialization step.
#define COMPLEX_INTITIALIZATION 1

// Complex initialization currently not supported with Verlet integration
#if VERLET_INTEGRATION
	#undef COMPLEX_INTITIALIZATION
	#define COMPLEX_INTITIALIZATION 0
#endif

// Whether to use EXT_pixel_buffer_object instead of NV_pixel_data_range
#define USE_PBO 1

CGcontext g_context;

void cgErrorCallback(void)
{
	CGerror lastError = cgGetError();
	if(lastError) {
		const char *listing = cgGetLastListing(g_context);
		printf("\n---------------------------------------------------\n");
		printf("%s\n\n", cgGetErrorString(lastError));
		printf("%s\n", listing);
		printf("-----------------------------------------------------\n");
		printf("Cg error, exiting...\n");

#ifdef _WIN32
		OutputDebugString("Cg Error:");
		OutputDebugString(cgGetErrorString(lastError));
		OutputDebugString("\n\n");
		OutputDebugString(listing);
		OutputDebugString("\n\n");
#endif

		cgDestroyContext(g_context);
		g_context = NULL;
		exit(1);
	}
}

ParticleSystem::ParticleSystem()
{
	PositionCurrent = 0;
	PositionDest = 1;
	PositionPrevious = 2;

	VelocityCurrent = 0;
	VelocityDest = 1;

	SortCurrent = 0;
	SortDest = 1;

	SortCurrentPass = 0;
	SortPassesPerFrame = 5;
	SortPassBegin = 0;
	SortPassEnd = SortPassesPerFrame;
	SortMaxPasses = 0;

	MaxUsedIndex = 0;

	TimeNow = 0.0f;
	LastInvalidateDeadTime = -1.0e10f;

	SpherePosition = float3(1.0f, 1.0f, 0.0f);

	FieldOfViewY = (float)M_PI_2;
	ScreenHeight = 500;
	ViewerPosition = float3(0.0f, -2.0f, -5.0f);
	ViewerRotation = float2(0.0f, 0.0f);

	// Init Cg.
	cgSetErrorCallback(cgErrorCallback);
	g_context = cgCreateContext();

	// Init configurable stuff.
	LoadConfigSetup("");

}

ParticleSystem::~ParticleSystem()
{
	UnloadConfigSetup();

	cgDestroyContext(g_context);
}

RenderTexture* ParticleSystem::CreateRenderTexture(int numOfBits, int numOfChannels)
{
	RenderTexture* renderTexture = new RenderTexture(Width,Height);
	int bits[4];
	for (int i = 0; i < 4; i++)
		bits[i] = i < numOfChannels ? numOfBits : 0;

	if (!renderTexture->Initialize(true, false, false, false, false, bits[0], bits[1], bits[2], bits[3]))
		exit(-1);

	renderTexture->BeginCapture();

	glDisable(GL_DITHER);
	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, Width, 0.0, Height, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0, 0, Width, Height);

	renderTexture->EndCapture();

	// create texture
	const GLenum target = GL_TEXTURE_RECTANGLE_NV;
	renderTexture->Bind();
	glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	return renderTexture;
}

CGprogram ParticleSystem::loadProgram(CGcontext context, CGprofile profile, char *filename, data_path &path, char* entryPoint /* = NULL */)
{
	std::string pathname = path.get_file(filename);
	if (pathname == "") {
		printf("Unable to load '%s', exiting...\n", filename);
		exit(-1);
	}

	CGprogram program = cgCreateProgramFromFile(context, CG_SOURCE, pathname.data(),
		profile, entryPoint, NULL);
	cgGLLoadProgram(program);
	return program;
}

CGprogram loadObjectProgram(CGcontext context, CGprofile profile, char *filename, data_path &path)
{
	std::string pathname = path.get_file(filename);
	if (pathname == "") {
		printf("Unable to load '%s', exiting...\n", filename);
		exit(-1);
	}

	CGprogram program = cgCreateProgramFromFile(context, CG_OBJECT, pathname.data(),
		profile, NULL, NULL);
	cgGLLoadProgram(program);
	return program;
}

void ParticleSystem::LoadConfigSetup(const string& setupPrefix, bool allowResize /* = true */)
{
	int oldWidth = Width;
	int oldHeight = Height;
	float3 oldViewerPosition = ViewerPosition;
	float2 oldViewerRotation = ViewerRotation;

	// Init optional values.
	FlowFieldTextureName = "flow_field.png";
	FlowFieldScale = float3(4, 1, 4);
	ClearColor = float4(0, 0, 0, 1);
	BlendFunction = BF_ADD;
	DoSorting = false;
	MenuKeys[0] = true;

	// Read config.
	ParticleSystemConfig config(Width, Height,
			Emitters, ParticleAttributes, DiffuseTextureName,
			LocalForcePositions,
			FlowFieldTextureName, FlowFieldScale,
			SpherePosition, ViewerPosition, ViewerRotation,
			ClearColor, BlendFunction, DoSorting, MenuKeys);

	if (setupPrefix.empty())
		config.ReadConfig("Config.xml");
	else
		config.ReadConfig(setupPrefix + "/Config.xml");

	if (!allowResize)
	{
		Width = oldWidth;
		Height = oldHeight;

		// Deactivate sorting if no buffers exist.
		if (!SortPBuffers[0])
			DoSorting = false;

		// To allow nice smooth transitions in this mode, disable moving the viewer around.
		ViewerPosition = oldViewerPosition;
		ViewerRotation = oldViewerRotation;
	}

	if (allowResize)
	{
		// Create RenderTexture objects.
		PositionBuffer = new float3[Width * Height];
		DebugBuffer = new float[Width * Height];

		PositionPBuffers[0] = CreateRenderTexture(32, 4);
		PositionPBuffers[1] = CreateRenderTexture(32, 4);
	#if VERLET_INTEGRATION
		PositionPBuffers[2] = CreateRenderTexture(32, 4);
		VelocityPBuffers[0] = NULL;
		VelocityPBuffers[1] = NULL;
	#else
		PositionPBuffers[2] = NULL;
		VelocityPBuffers[0] = CreateRenderTexture(16, 4);
		VelocityPBuffers[1] = CreateRenderTexture(16, 4);
	#endif

		DeathTimePBuffer = CreateRenderTexture(32, 1);

		if (DoSorting)
		{
			SortPBuffers[0] = CreateRenderTexture(32, 2);
			SortPBuffers[1] = CreateRenderTexture(32, 2);
		}
		else
		{
			SortPBuffers[0] = NULL;
			SortPBuffers[1] = NULL;
		}

		// Init vertex arrays.
#if USE_PBO
		RenderVertexBuffer = new RenderVertexArray(0, GetSize(), 4); // 0==position

		ParticleVertices = new ParticleVertexData[GetSize()];
		memset(ParticleVertices, sizeof(ParticleVertices[0]) * GetSize(), 0);

		// create vbo for particle vertex data
		glGenBuffersARB(1, &ParticleVertexBuffer);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, ParticleVertexBuffer);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, GetSize() * 4 * sizeof(float), 0, GL_DYNAMIC_DRAW);
		glVertexAttribPointerARB(8, 4, GL_FLOAT, GL_FALSE, 0, 0);  // 8 = texcoord0
#else
		CurrentVertexArray = 0;
		int i;
		for (i = 0; i < sizeof(VertexArray) / sizeof(VertexArray[0]); i++)
			VertexArray[i] = new RenderableVertexArray(Width * Height, 7);
#endif
	}

	string prefix;
	if (!setupPrefix.empty())
		prefix = "/" + setupPrefix;

	data_path path;
	path.path.push_back(".");
	path.path.push_back("./shaders" + prefix);
	path.path.push_back("../media/shaders" + prefix);
	path.path.push_back("../../media/shaders" + prefix);


	// load programs
	MovementVerletFProg = loadProgram(g_context, CG_PROFILE_FP30, "moveVerlet.cg", path);

	MovementFProg = loadProgram(g_context, CG_PROFILE_FP30, "move.cg", path);
	MovementInitFProg = loadProgram(g_context, CG_PROFILE_FP30, "move.cg", path, "mainInitialization");

	VelocityFProg = loadProgram(g_context, CG_PROFILE_FP30, "velocity.cg", path);
	VelocityInitFProg = loadProgram(g_context, CG_PROFILE_FP30, "velocity.cg", path, "mainInitialization");

	PassthruFProg = loadProgram(g_context, CG_PROFILE_FP30, "passthru.cg", path);
	InitValueFProg = loadProgram(g_context, CG_PROFILE_FP30, "init_value.cg", path);
	InvalidateDeadFProg = loadProgram(g_context, CG_PROFILE_FP30, "invalidate_dead.cg", path);

	SortRecursionFProg = loadProgram(g_context, CG_PROFILE_FP30, "mergesort.cg", path, "mergeSortRecursion");
	SortEndFProg = loadProgram(g_context, CG_PROFILE_FP30, "mergesort.cg", path, "mergeSortEnd");

	DistanceSortInitFProg = loadProgram(g_context, CG_PROFILE_FP30, "distancesort.cg", path, "initSortIndex");
	DistanceSortComputeFProg = loadProgram(g_context, CG_PROFILE_FP30, "distancesort.cg", path, "computeDistance");
	DistanceSortLookupFProg = loadProgram(g_context, CG_PROFILE_FP30, "distancesort.cg", path, "lookupPosition");

	DisplayFProg = loadProgram(g_context, CG_PROFILE_FP30, "display_fp.cg", path);
	DisplayVProg = loadProgram(g_context, CG_PROFILE_VP30, "display_vp.cg", path);
	//DisplayVProg = loadObjectProgram(g_context, CG_PROFILE_VP30, "shader_vp.vp30cg", path);
	//fprintf(stderr, "---- PROGRAM BEGIN ----\n%s---- PROGRAM END ----\n",
	//		cgGetProgramString(DisplayVProg, CG_COMPILED_PROGRAM));

	// get parameters
	MovementVerlet_TimeStep_Param = cgGetNamedParameter(MovementVerletFProg, "_TimeStep");
	MovementVerlet_SphereCenter_Param = cgGetNamedParameter(MovementVerletFProg, "_SphereCenter");
	//CGparameter meshSize_Param = cgGetNamedParameter(MovementVerletFProg, "_MeshSize");
	//cgGLSetParameter2f(meshSize_Param, (float)(Width - 1), (float)(Height - 1));

	Movement_TimeStep_Param = cgGetNamedParameter(MovementFProg, "_TimeStep");

	Velocity_TimeStep_Param = cgGetNamedParameter(VelocityFProg, "_TimeStep");
	Velocity_SphereCenter_Param = cgGetNamedParameter(VelocityFProg, "_SphereCenter");
	Velocity_LocalAttractorPosition_Param = cgGetNamedParameter(VelocityFProg, "_LocalAttractorPosition");

	VelocityInit_SphereCenter_Param = cgGetNamedParameter(VelocityInitFProg, "_SphereCenter");
	VelocityInit_LocalAttractorPosition_Param = cgGetNamedParameter(VelocityInitFProg, "_LocalAttractorPosition");

	Passthru_Scale_Param = cgGetNamedParameter(PassthruFProg, "_Scale");
	Passthru_Bias_Param = cgGetNamedParameter(PassthruFProg, "_Bias");

	InvalidateDead_CurrentTime_Infinity_Param = cgGetNamedParameter(InvalidateDeadFProg, "_CurrentTime_Infinity");

	SortRecursion_Size_Param = cgGetNamedParameter(SortRecursionFProg, "_Size");
	SortRecursion_Step_Param = cgGetNamedParameter(SortRecursionFProg, "_Step");
	SortRecursion_Count_Param = cgGetNamedParameter(SortRecursionFProg, "_Count");
	SortEnd_Size_Param = cgGetNamedParameter(SortEndFProg, "_Size");
	SortEnd_Step_Param = cgGetNamedParameter(SortEndFProg, "_Step");

	DistanceSortCompute_ViewerPosition_Param = cgGetNamedParameter(DistanceSortComputeFProg, "_ViewerPosition");

	DisplayVP_ModelViewProj_Param = cgGetNamedParameter(DisplayVProg, "_ModelViewProj");
	DisplayVP_ModelView_Param = cgGetNamedParameter(DisplayVProg, "_ModelView");
	DisplayVP_TanHalfFOVTimesScreenSize_Param = cgGetNamedParameter(DisplayVProg, "_TanHalfFOV");
	DisplayVP_CurrentTime_Param = cgGetNamedParameter(DisplayVProg, "_CurrentTime");
	DisplayVP_RandomTable_Param = cgGetNamedParameter(DisplayVProg, "_RandomTable");

	ParticleAttributesParameters.resize(GetNumOfAttributeTypes());
	int i;
	for (i = 0; i < GetNumOfAttributeTypes(); i++)
	{
		stringstream sstr;
		sstr << "[" << i << "]";
		ParticleAttributesParameters[i].PointSizeRange = cgGetNamedParameter(DisplayVProg, ("_ParticleAttributes_PointSizeRange" + sstr.str()).c_str());
		ParticleAttributesParameters[i].OrientationStartRange_VelocityRange = cgGetNamedParameter(DisplayVProg, ("_ParticleAttributes_OrientationStartRange_VelocityRange" + sstr.str()).c_str());
		ParticleAttributesParameters[i].TileScale_TileOffset = cgGetNamedParameter(DisplayVProg, ("_ParticleAttributes_TileScale_TileOffset" + sstr.str()).c_str());
		ParticleAttributesParameters[i].ColorAnimationSegments = cgGetNamedParameter(DisplayVProg, ("_ParticleAttributes_ColorAnimationSegments" + sstr.str()).c_str());
		ParticleAttributesParameters[i].ColorAnimationFunctions = cgGetNamedParameter(DisplayVProg, ("_ParticleAttributes_ColorAnimationFunctions" + sstr.str()).c_str());
	}

	data_path texturePath;
	texturePath.path.push_back("");
	texturePath.path.push_back(".");
	texturePath.path.push_back("./textures");
	texturePath.path.push_back("../media/textures");
	texturePath.path.push_back("../../media/textures");
	set_png_path(texturePath);

	// Load diffuse texture.
	array2<vec3ub> img;
	array2<vec4ub> alphaImg;
	nv_dds::CDDSImage ddsImg;
	ddsImg.load(texturePath.get_file(DiffuseTextureName));

	glGenTextures(1, &DiffuseTexture);
	glBindTexture(GL_TEXTURE_2D, DiffuseTexture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	ddsImg.upload_texture2D(0, GL_TEXTURE_2D, GL_RGBA8);

	// Load terrain texture.
	read_png_grey("terrain.png", TerrainHeightField);
	glGenTextures(1, &TerrainTexture);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, TerrainTexture);
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_LUMINANCE8,
		TerrainHeightField.get_width(), TerrainHeightField.get_height(), 0,
		GL_LUMINANCE, GL_UNSIGNED_BYTE, (const void *)TerrainHeightField.get_pointer());
	TerrainSimpleMesh.CreateFromHeightField(TerrainHeightField);

	// Load flow texture.
	read_png_rgb(FlowFieldTextureName.c_str(), img);
	FlowFieldSize[0] = img.get_width();
	FlowFieldSize[1] = img.get_height();
	glGenTextures(1, &FlowFieldTexture);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, FlowFieldTexture);
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGB8,
		img.get_width(), img.get_height(), 0,
		GL_RGB, GL_UNSIGNED_BYTE, (const void *)img.get_pointer());

	// Init random table.
	if (DisplayVP_RandomTable_Param)
	{
		RandomTable.resize(cgGetArraySize(DisplayVP_RandomTable_Param, 0));
		for (i = 0; i < (int)RandomTable.size(); i++)
			for (int j = 0; j < 4; j++)
				RandomTable[i][j] = randomRange(0.0f, 1.0f);
	}

}

void ParticleSystem::UnloadConfigSetup(bool allowResize /* = true */)
{
	glDeleteTextures(1, &DiffuseTexture);
	DiffuseTexture = 0;
	glDeleteTextures(1, &TerrainTexture);
	TerrainTexture = 0;
	glDeleteTextures(1, &FlowFieldTexture);
	FlowFieldTexture = 0;

	cgDestroyProgram(MovementVerletFProg);
	MovementVerletFProg = 0;
	cgDestroyProgram(MovementFProg);
	MovementFProg = 0;
	cgDestroyProgram(MovementInitFProg);
	MovementInitFProg = 0;

	cgDestroyProgram(VelocityFProg);
	VelocityFProg = 0;
	cgDestroyProgram(VelocityInitFProg);
	VelocityInitFProg = 0;

	cgDestroyProgram(PassthruFProg);
	PassthruFProg = 0;
	cgDestroyProgram(InitValueFProg);
	InitValueFProg = 0;
	cgDestroyProgram(InvalidateDeadFProg);
	InvalidateDeadFProg = 0;

	cgDestroyProgram(SortRecursionFProg);
	SortRecursionFProg = 0;
	cgDestroyProgram(SortEndFProg);
	SortEndFProg = 0;

	cgDestroyProgram(DistanceSortInitFProg);
	DistanceSortInitFProg = 0;
	cgDestroyProgram(DistanceSortComputeFProg);
	DistanceSortComputeFProg = 0;
	cgDestroyProgram(DistanceSortLookupFProg);
	DistanceSortLookupFProg = 0;

	cgDestroyProgram(DisplayFProg);
	DisplayFProg = 0;
	cgDestroyProgram(DisplayVProg);
	DisplayVProg = 0;

	Emitters.resize(0);
	ParticleAttributes.resize(0);
	LocalForcePositions.resize(0);

	if (allowResize)
	{
		delete [] PositionBuffer;
		PositionBuffer = 0;
		delete [] DebugBuffer;
		DebugBuffer = 0;

		int i;
		for (i = 0; i < sizeof(PositionPBuffers) / sizeof(PositionPBuffers[0]); i++)
		{
			delete PositionPBuffers[i];
			PositionPBuffers[i] = 0;
		}
		for (i = 0; i < sizeof(VelocityPBuffers) / sizeof(VelocityPBuffers[0]); i++)
		{
			delete VelocityPBuffers[i];
			VelocityPBuffers[i] = 0;
		}
		delete DeathTimePBuffer;
		DeathTimePBuffer = 0;
		for (i = 0; i < sizeof(SortPBuffers) / sizeof(SortPBuffers[0]); i++)
		{
			delete SortPBuffers[i];
			SortPBuffers[i] = 0;
		}
#if USE_PBO
		delete RenderVertexBuffer;
		RenderVertexBuffer = NULL;

		delete [] ParticleVertices;
		ParticleVertices = NULL;

		// create vbo for particle vertex data
		glDeleteBuffersARB(1, &ParticleVertexBuffer);
		ParticleVertexBuffer = NULL;
#else
		for (int i = 0; i < sizeof(VertexArray) / sizeof(VertexArray[0]); i++)
		{
			delete VertexArray[i];
			VertexArray[i] = 0;
		}
		CurrentVertexArray = 0;
#endif
	}

}

void ParticleSystem::DrawImage()
{
	cgGLBindProgram(PassthruFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	cgGLSetParameter4f(Passthru_Scale_Param, 1.0, 1.0, 1.0, 1.0);
	cgGLSetParameter4f(Passthru_Bias_Param, 0.0, 0.0, 0.0, 0.0);

	//    glRasterPos3f(0, 0, 0);
	//    glDrawPixels(Width, Height, GL_RGB, GL_FLOAT, PositionBuffer);

	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_FLOAT_RGB_NV, Width, Height, 0, GL_RGB, GL_FLOAT, (GLfloat *) &PositionBuffer[0]);
	DrawQuad();

	cgGLDisableProfile(CG_PROFILE_FP30);
}

// Initialize particles.
void ParticleSystem::Reset()
{
	while (!AvailableIndices.empty())
		AvailableIndices.pop();
	while (!ActiveParticles.empty())
		ActiveParticles.pop();

	AvailableIndices.reserve(GetSize());
	UsedIndices.clear();
	UsedIndices.resize(GetSize(), false);
	ActiveParticles.reserve(GetSize());

	int i;
	//for (i = GetSize() - 1; i >= 0; i--)
	for (i = 0; i < GetSize(); i++)
		AvailableIndices.push(i);

#if USE_PBO
	ParticleVertexData* pvd = ParticleVertices;
#else
	ParticleVertexData* pvd = (ParticleVertexData*)VertexArray[CurrentVertexArray]->GetPointerBehindPositions();
#endif
	for (i = 0; i < GetSize(); i++)
	{
		pvd[i].TimeOfBirth = -1.0f;
		pvd[i].LifeTime = 0.0f;
		pvd[i].AttributeType = 0.0f;
		pvd[i].Index = (float)i;
	}

	for(int y = 0; y < Height; y++) {
		for(int x = 0; x < Width; x++) {
			//float3 p = (float3((float)rand() / RAND_MAX, (float)rand() / RAND_MAX, (float)rand() / RAND_MAX)
			//	- float3(0.5, 0, 0.5)) * 4 ;
			float3 p(numeric_limits<float>::max());
			PositionBuffer[Width * y + x] = p;
		}
	}
	PositionPBuffers[PositionCurrent]->BeginCapture();
	glClear(GL_COLOR_BUFFER_BIT);
	DrawImage();
	PositionPBuffers[PositionCurrent]->EndCapture();

#if VERLET_INTEGRATION
	PositionPBuffers[PositionPrevious]->BeginCapture();
	glClear(GL_COLOR_BUFFER_BIT);
	DrawImage();
	PositionPBuffers[PositionPrevious]->EndCapture();
//#if VERLET_INTEGRATION
#else
	VelocityPBuffers[VelocityCurrent]->BeginCapture();
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	VelocityPBuffers[VelocityCurrent]->EndCapture();
#endif

	DeathTimePBuffer->BeginCapture();
	glClear(GL_COLOR_BUFFER_BIT);
	DeathTimePBuffer->EndCapture();

	if (DoSorting)
	{
		//srand(157);
		//srand((unsigned)time(NULL));
		vector<float> sortData(GetSize());
		for (int i = 0; i < (int)sortData.size(); i++)
			sortData[i] = (float)i / (sortData.size() - 1);
		for (int i = 0; i < (int)sortData.size(); i++)
		{
			int p = (int)(rand() * (__int64)(sortData.size() - 1) / RAND_MAX);
			float f = sortData[i];
			sortData[i] = sortData[p];
			sortData[p] = f;
		}

		for(int y = 0; y < Height; y++) {
			for(int x = 0; x < Width; x++) {
				PositionBuffer[Width * y + x] = float3(sortData[Width * y + x], sortData[Width * y + x], sortData[Width * y + x]);
			}
		}
		
		SortPBuffers[SortCurrent]->BeginCapture();
		glClear(GL_COLOR_BUFFER_BIT);
		DrawImage();
		SortPBuffers[SortCurrent]->EndCapture();

		SortCurrentPass = 0;
		SortMaxPasses = 100000;



		SortPBuffers[SortCurrent]->BeginCapture();

		cgGLBindProgram(DistanceSortInitFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		SortPBuffers[SortCurrent]->EndCapture();


		/*
		SortPBuffers[SortDest]->BeginCapture();
		glReadPixels(0, 0, Width, Height, GL_RGB, GL_FLOAT, (float*)PositionBuffer);
		SortPBuffers[SortDest]->EndCapture();

		for (int i = 0; i < GetSize(); i++)
			printf("%.3f, %.3f\n", PositionBuffer[i].v[0], PositionBuffer[i].v[1]);
		printf("\n");
		*/
	}

}

void ParticleSystem::SetViewerPosition(const float3& position)
{
	ViewerPosition = position;
}

const float3& ParticleSystem::GetViewerPosition()
{
	return ViewerPosition;
}

void ParticleSystem::SetViewerRotation(const float2& rotation)
{
	ViewerRotation = rotation;
}

const float2& ParticleSystem::GetViewerRotation()
{
	return ViewerRotation;
}

void ParticleSystem::SetSpherePosition(const float3& position)
{
	SpherePosition = position;
}

const float3& ParticleSystem::GetSpherePosition()
{
	return SpherePosition;
}

void ParticleSystem::SetEmitterPosition(int emitterIndex, const float3& position)
{
	Emitters[emitterIndex].Position = position;
}

const float3& ParticleSystem::GetEmitterPosition(int emitterIndex)
{
	return Emitters[emitterIndex].Position;
}

void ParticleSystem::SetEmitterDirection(int emitterIndex, const float3& direction)
{
	Emitters[emitterIndex].InitialDirectionCenter = direction;
}

const float3& ParticleSystem::GetEmitterDirection(int emitterIndex)
{
	return Emitters[emitterIndex].InitialDirectionCenter;
}

void ParticleSystem::SetLocalForcePosition(int localForceIndex, const float3& position)
{
	LocalForcePositions[localForceIndex] = position;
}

const float3& ParticleSystem::GetLocalForcePosition(int localForceIndex)
{
	return LocalForcePositions[localForceIndex];
}

void ParticleSystem::SetPerspective(float fieldOfViewY, int screenHeight)
{
	FieldOfViewY = fieldOfViewY;
	ScreenHeight = screenHeight;
}


// step the simulation
void ParticleSystem::TimeStep(float dT)
{
	// Update depth map of dynamic collider.
	//DepthMapCollider->generateDepthTextures();

#if VERLET_INTEGRATION
	// Integration pass.
	// read from current and previous position buffers, write to destination
	PositionPBuffers[PositionDest]->BeginCapture();

	cgGLBindProgram(MovementVerletFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);
	
	glActiveTextureARB(GL_TEXTURE0_ARB);
	PositionPBuffers[PositionCurrent]->Bind();

	glActiveTextureARB(GL_TEXTURE1_ARB);
	PositionPBuffers[PositionPrevious]->Bind();

	cgGLSetParameter1f(MovementVerlet_TimeStep_Param, dT);
	if (MovementVerlet_SphereCenter_Param)
		cgGLSetParameter3fv(MovementVerlet_SphereCenter_Param, SpherePosition.v);

	DrawQuad();

	cgGLDisableProfile(CG_PROFILE_FP30);

	PositionPBuffers[PositionDest]->EndCapture();

	int temp = PositionPrevious;
	PositionPrevious = PositionCurrent;
	PositionCurrent = PositionDest;
	PositionDest = temp;

#else // VERLET_INTEGRATION

	// Velocity change pass.
	// Read from position and velocity buffers, write to destination
	VelocityPBuffers[VelocityDest]->BeginCapture();

	cgGLBindProgram(VelocityFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	VelocityPBuffers[VelocityCurrent]->Bind();

	glActiveTextureARB(GL_TEXTURE1_ARB);
	PositionPBuffers[PositionCurrent]->Bind();

	/*glActiveTextureARB(GL_TEXTURE2_ARB);
	PositionPBuffers[PositionPrevious]->Bind();*/

	glActiveTextureARB(GL_TEXTURE2_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, TerrainTexture);

	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, FlowFieldTexture);

	cgGLSetParameter1f(Velocity_TimeStep_Param, dT);
	if (Velocity_SphereCenter_Param)
		cgGLSetParameter3fv(Velocity_SphereCenter_Param, SpherePosition.v);
	if (Velocity_LocalAttractorPosition_Param)
		cgGLSetParameterArray3f(Velocity_LocalAttractorPosition_Param, 0, GetNumOfLocalForces(), LocalForcePositions[0].v);

	DrawQuad();

	cgGLDisableProfile(CG_PROFILE_FP30);

	VelocityPBuffers[VelocityDest]->EndCapture();

	int temp = VelocityCurrent;
	VelocityCurrent = VelocityDest;
	VelocityDest = temp;

	// Movement integration pass using Euler integration.
	// Read from position and velocity buffers, write to destination
	PositionPBuffers[PositionDest]->BeginCapture();

	cgGLBindProgram(MovementFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	PositionPBuffers[PositionCurrent]->Bind();

	glActiveTextureARB(GL_TEXTURE1_ARB);
	VelocityPBuffers[VelocityCurrent]->Bind();

	cgGLSetParameter1f(Movement_TimeStep_Param, dT);

	DrawQuad();

	cgGLDisableProfile(CG_PROFILE_FP30);

	PositionPBuffers[PositionDest]->EndCapture();

	temp = PositionCurrent;
	PositionCurrent = PositionDest;
	PositionDest = temp;
	/*temp = PositionPrevious;
	PositionPrevious = PositionCurrent;
	PositionCurrent = PositionDest;
	PositionDest = temp;*/
#endif

// Currently there's a flicker bug when use PBO with the cleanup pass. So disable it for now.
// The cleanup is not completely neccessary, since particles fade out at the end
// of their life time anyway, but moving them outside the viewing frustum might improve performance.
#if !USE_PBO
	if (TimeNow > LastInvalidateDeadTime + 10)
	{
		// Invalidate dead particles pass.
		PositionPBuffers[PositionDest]->BeginCapture();

		cgGLBindProgram(InvalidateDeadFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		PositionPBuffers[PositionCurrent]->Bind();

		glActiveTextureARB(GL_TEXTURE1_ARB);
		DeathTimePBuffer->Bind();

		cgGLSetParameter2f(InvalidateDead_CurrentTime_Infinity_Param, TimeNow, numeric_limits<float>::infinity());

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		PositionPBuffers[PositionDest]->EndCapture();

		temp = PositionCurrent;
		PositionCurrent = PositionDest;
		PositionDest = temp;

		LastInvalidateDeadTime = TimeNow;
	}
#endif


	// Manage allocation/deallocation.
	for (int e = 0; e < GetNumOfEmitters(); e++)
	{
		int numOfNewParticles = (int)(dT * randomRange(Emitters[e].NumOfNewParticlesMinMax[0], Emitters[e].NumOfNewParticlesMinMax[1]));
			// (int)((1.0f - (TimeNow / 200.0f - floorf(TimeNow / 200.0f))) * 200.0f);

		for (int i = 0; i < numOfNewParticles; i++)
		{
			float3 position = Emitters[e].PreviousPosition * (1 - (float)i / numOfNewParticles)
				+ Emitters[e].Position * ((float)i / numOfNewParticles);

			position += Emitters[e].PositionMaxOffset * float3(randomRange(-1.0f, 1.0f), randomRange(-1.0f, 1.0f), randomRange(-1.0f, 1.0f));

			//float speed = 0.05f;
			float speed = randomRange(Emitters[e].InitialVelocityMinMax[0], Emitters[e].InitialVelocityMinMax[1]);

			//float3 direction(0, -1, 0);
			//float3 direction(randomRange(-1.0f, 1.0f), -3.0f, randomRange(-1.0f, 1.0f));
			float3 direction(Emitters[e].InitialDirectionCenter);

			float3 ortho(-direction[1], direction[0], 0.0f);
			if (ortho.length() < 0.01f)
				ortho = float3(-direction[2], 0.0f, direction[0]);
			ortho.normalize();

#if 0
			ortho *= tan(randomRange(0.0f, Emitters[e].InitialDirectionMaxAngle));
			float angle = randomRange(0.0f, 2.0f * (float)M_PI);
			direction = direction + sin(angle) * ortho + cos(angle) * direction.cross(ortho);
#else
			float2 randomDisc;
			int j;
			for (j = 0; j < 5; j++)
			{
				randomDisc[0] = randomRange(-1.0f, 1.0f);
				randomDisc[1] = randomRange(-1.0f, 1.0f);
				if (randomDisc.length() <= 1.0f)
					break;
			}
			if (j == 5)
				randomDisc.normalize();
			randomDisc *= tan(Emitters[e].InitialDirectionMaxAngle);

			direction = direction + randomDisc[0] * ortho + randomDisc[1] * direction.cross(ortho);
#endif
			direction.normalize();

			float3 velocity = direction * speed;

			//position += velocity * (1 - (float)i / numOfNewParticles);

			float lifeTime = randomRange(Emitters[e].LifeTimeMinMax[0], Emitters[e].LifeTimeMinMax[1]);

			AllocateParticle(position, velocity, lifeTime, Emitters[e].AttributeType, (1.0f - (float)i / numOfNewParticles));
		}

		Emitters[e].PreviousPosition = Emitters[e].Position;
	}

	InitNewParticles();

	TimeNow += dT;

	CleanUpParticles();

	//Sleep(200);
}

void ParticleSystem::DrawQuad()
{
	glBegin(GL_QUADS);

	glTexCoord2f(0, 0);
	glVertex2f(0, 0);

	glTexCoord2f((float)Width, 0);
	glVertex2f((float)Width, 0);

	glTexCoord2f((float)Width, (float)Height); 
	glVertex2f((float)Width, (float)Height);

	glTexCoord2f(0, (float)Height);
	glVertex2f(0, (float)Height);

	glEnd();
}

// read frame buffer to main memory
void ParticleSystem::ReadBack()
{
	PositionPBuffers[PositionCurrent]->BeginCapture();
	glReadPixels(0, 0, Width, Height, GL_RGB, GL_FLOAT, (float*) PositionBuffer);
	//glReadPixels(0, 0, Width, Height, GL_ALPHA, GL_FLOAT, (float*) DebugBuffer);
	PositionPBuffers[PositionCurrent]->EndCapture();
}

// display points using render-to-vertex array
void ParticleSystem::DisplayPoints()
{

	int positionRender; // Buffer index to render from.

	if (DoSorting)
	{
		// Update distances in sort texture.
		SortPBuffers[SortDest]->BeginCapture();

		cgGLBindProgram(DistanceSortComputeFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		SortPBuffers[SortCurrent]->Bind();

		glActiveTextureARB(GL_TEXTURE1_ARB);
		PositionPBuffers[PositionCurrent]->Bind();

		printf("Viewer: %.2f %.2f %.2f\n", ViewerPosition.v[0], ViewerPosition.v[1], ViewerPosition.v[2]);
		cgGLSetParameter3fv(DistanceSortCompute_ViewerPosition_Param, ViewerPosition.v);

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		SortPBuffers[SortDest]->EndCapture();

		int temp = SortCurrent;
		SortCurrent = SortDest;
		SortDest = temp;

		// Sort.
		MergeSort();

		// Lookup positions from sorted texture.
		PositionPBuffers[PositionDest]->BeginCapture();

		cgGLBindProgram(DistanceSortLookupFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		SortPBuffers[SortCurrent]->Bind();

		glActiveTextureARB(GL_TEXTURE1_ARB);
		PositionPBuffers[PositionCurrent]->Bind();

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		PositionPBuffers[PositionDest]->EndCapture();

		// Don't swap PositionDest and PositionCurrent this time,
		// because simulation must not use sorted positions!

		positionRender = PositionDest;
	}
	else
		positionRender = PositionCurrent;

#if USE_PBO
	PositionPBuffers[positionRender]->BeginCapture();
 	RenderVertexBuffer->Read(GL_FRONT, Width, Height);
    PositionPBuffers[positionRender]->EndCaptureWithoutCopy();
#else
	int previousVA = CurrentVertexArray;
	CurrentVertexArray = (CurrentVertexArray + 1) % (sizeof(VertexArray) / sizeof(VertexArray[0]));

	VertexArray[CurrentVertexArray]->Read(PositionPBuffers[positionRender]);
#endif

	cgGLBindProgram(DisplayFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	cgGLBindProgram(DisplayVProg);
	cgGLEnableProfile(CG_PROFILE_VP30);

	cgGLSetStateMatrixParameter(DisplayVP_ModelViewProj_Param, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	cgGLSetStateMatrixParameter(DisplayVP_ModelView_Param, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
	cgGLSetParameter1f(DisplayVP_TanHalfFOVTimesScreenSize_Param, tan(FieldOfViewY / 2.0f) * ScreenHeight);
	cgGLSetParameter1f(DisplayVP_CurrentTime_Param, TimeNow);
	cgGLSetParameterArray4f(DisplayVP_RandomTable_Param, 0, RandomTable.size(), RandomTable[0].v);
	
	for (int i = 0; i < GetNumOfAttributeTypes(); i++)
	{
		cgGLSetParameter2fv(ParticleAttributesParameters[i].PointSizeRange, ParticleAttributes[i].PointSizeRange.v);
		cgGLSetParameter4fv(ParticleAttributesParameters[i].OrientationStartRange_VelocityRange, ParticleAttributes[i].OrientationStartRange_VelocityRange.v);
		cgGLSetParameter4fv(ParticleAttributesParameters[i].TileScale_TileOffset, ParticleAttributes[i].TileScale_TileOffset.v);
		cgGLSetParameter4fv(ParticleAttributesParameters[i].ColorAnimationSegments, ParticleAttributes[i].ColorAnimationSegments.v);
		cgGLSetParameterArray4f(ParticleAttributesParameters[i].ColorAnimationFunctions, 0, 6, ParticleAttributes[i].ColorAnimationFunctions[0].v);
	}

	glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
	if (BlendFunction == BF_ADD)
		glBlendFunc(GL_ONE, GL_ONE);
	else if (BlendFunction == BF_ALPHA)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	else if (BlendFunction == BF_PREMULTIPLIED_ALPHA)
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glColor3f(1.0f, 0.1f, 0.03f);

	glEnable(GL_POINT_SPRITE_NV);

	//glPointSize(10.0f);
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE_NV);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_2D, DiffuseTexture);
	glTexEnvi(GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_TRUE);

	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

#if USE_PBO
	glEnableVertexAttribArrayARB(8);
 	RenderVertexBuffer->SetPointer();
    glDrawArrays(GL_POINTS, 0, GetSize());
#else
	glTexCoordPointer(4, GL_FLOAT, 0, VertexArray[previousVA]->GetPointerBehindPositions());
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	VertexArray[previousVA]->SetPointers();

	//VertexArray[previousVA]->DrawArrays(GL_POINTS, 0, MaxUsedIndex);
	VertexArray[previousVA]->DrawArrays(GL_POINTS, 0, GetSize());
#endif

	/*glBegin(GL_POINTS);
	glVertex3f(0, 1, 0);
	glEnd();*/

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

#if USE_PBO
    glDisableVertexAttribArrayARB(0);
    glDisableVertexAttribArrayARB(8);
#endif

	glPopClientAttrib();

	glTexEnvi(GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_FALSE);
	glDisable(GL_VERTEX_PROGRAM_POINT_SIZE_NV);
	glDisable(GL_POINT_SPRITE_NV);
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);

	cgGLDisableProfile(CG_PROFILE_VP30);
	cgGLDisableProfile(CG_PROFILE_FP30);
}

// display float texture using texturing
void ParticleSystem::DisplayTexture(RenderTexture* renderTexture)
{
	cgGLBindProgram(PassthruFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	renderTexture->Bind();
	DrawQuad();
	cgGLDisableProfile(CG_PROFILE_FP30);
}

// display position and normal textures
void ParticleSystem::DisplayTextures()
{
	glPushMatrix();
	glScalef(0.01f, 0.01f, 1.0f);
	glTranslatef((float)-Width/2, (float)-Height, 0.0);
	cgGLSetParameter4f(Passthru_Scale_Param, 0.25, 0.25, 0.25, 1.0);
	cgGLSetParameter4f(Passthru_Bias_Param, 0.5, 0.0, 0.5, 0.0);
	DisplayTexture(PositionPBuffers[PositionCurrent]);
	glPopMatrix();

#if VERLET_INTEGRATION
	glPushMatrix();
	glScalef(0.01f, 0.01f, 1.0f);
	glTranslatef((float)Width/2, (float)-Height, 0.0);
	cgGLSetParameter4f(Passthru_Scale_Param, 0.25, 0.25, 0.25, 1.0);
	cgGLSetParameter4f(Passthru_Bias_Param, 0.5, 0.0, 0.5, 0.0);
	DisplayTexture(PositionPBuffers[PositionPrevious]);
	glPopMatrix();
#else
	glPushMatrix();
	glScalef(0.01f, 0.01f, 1.0f);
	glTranslatef((float)Width/2, (float)-Height, 0.0);
	cgGLSetParameter4f(Passthru_Scale_Param, 4.0, 4.0, 4.0, 1.0);
	cgGLSetParameter4f(Passthru_Bias_Param, 0.5, 0.5, 0.5, 0.0);
	DisplayTexture(VelocityPBuffers[VelocityCurrent]);
	glPopMatrix();
#endif

	if (DoSorting)
	{
		glPushMatrix();
		glScalef(0.01f, 0.01f, 1.0f);
		glTranslatef((float)-Width/2, 0.0, 0.0);
		cgGLSetParameter4f(Passthru_Scale_Param, 0.025f, 0.0, 0.0, 1.0);
		cgGLSetParameter4f(Passthru_Bias_Param, 0.0, 0.0, 0.0, 0.0);
		DisplayTexture(SortPBuffers[SortCurrent]);
		glPopMatrix();
	}
}

void ParticleSystem::AllocateParticle(const float3& position, const float3& velocity,
		float lifeTime, int attributeType, float initTimeStep)
{
	//assert(!AvailableIndices.empty());
	if (AvailableIndices.empty())
		return;

	int i = AvailableIndices.top();
	AvailableIndices.pop();

	ActiveParticle p;
	p.Index = i;
	p.TimeOfDeath = TimeNow + lifeTime;

	ActiveParticles.push(p);

	UsedIndices[i] = true;
	if (i >	MaxUsedIndex)
		MaxUsedIndex = i;

	NewParticle n;
	n.Index = i;
	n.Position = position;
	n.Velocity = velocity;
	n.TimeOfDeath = p.TimeOfDeath;
	n.InitTimeStep = initTimeStep;

	NewParticles.push_back(n);

#if USE_PBO
	ParticleVertexData* pvd = ParticleVertices;
#else
	ParticleVertexData* pvd = (ParticleVertexData*)VertexArray[CurrentVertexArray]->GetPointerBehindPositions();
#endif

	pvd[i].TimeOfBirth = TimeNow;
	pvd[i].LifeTime = lifeTime;
	pvd[i].AttributeType = (float)attributeType;

}

void ParticleSystem::InitNewParticles()
{
	vector<NewParticle>::iterator i;

#ifdef _WIN32
	// This offset is used to work around a driver inaccuracy when rendering pixels.
	static const int offsetY = 1;
#else
	static const int offsetY = 0;
#endif

#if USE_PBO
	// Upload new particle data to vertex buffer
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, ParticleVertexBuffer);
	ParticleVertexData* mappedParticleVertices = (ParticleVertexData*)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);

	for (i = NewParticles.begin(); i != NewParticles.end(); i++)
	{
		mappedParticleVertices[i->Index] = ParticleVertices[i->Index];
	}

	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);
#endif

#if VERLET_INTEGRATION
	// Render velocity (via fake previous position) into texture.
	PositionPBuffers[PositionPrevious]->BeginCapture();

	cgGLBindProgram(InitValueFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glBegin(GL_POINTS);
	for (i = NewParticles.begin(); i != NewParticles.end(); i++)
	{
		glTexCoord3fv((i->Position - i->Velocity).v);
		glVertex2f((float)(i->Index % Width),
			(float)(i->Index / Height) + offsetY /* + offsetY shouldn't be */);
	}
	glEnd();

	cgGLDisableProfile(CG_PROFILE_FP30);

	PositionPBuffers[PositionPrevious]->EndCapture();
#else
	// Render velocity.
	VelocityPBuffers[VelocityCurrent]->BeginCapture();

	#if COMPLEX_INTITIALIZATION
		cgGLBindProgram(VelocityInitFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE2_ARB);
		glBindTexture(GL_TEXTURE_RECTANGLE_NV, TerrainTexture);
		glActiveTextureARB(GL_TEXTURE3_ARB);
		glBindTexture(GL_TEXTURE_RECTANGLE_NV, FlowFieldTexture);

		if (VelocityInit_SphereCenter_Param)
			cgGLSetParameter2fv(VelocityInit_SphereCenter_Param, SpherePosition.v);
		if (VelocityInit_LocalAttractorPosition_Param)
			cgGLSetParameterArray3f(VelocityInit_LocalAttractorPosition_Param, 0, GetNumOfLocalForces(), LocalForcePositions[0].v);
	#else
		cgGLBindProgram(InitValueFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);
	#endif

	glBegin(GL_POINTS);
	for (i = NewParticles.begin(); i != NewParticles.end(); i++)
	{
		glMultiTexCoord3fv(0, i->Velocity.v);
	#if COMPLEX_INTITIALIZATION
			glMultiTexCoord3fv(1, i->Position.v);
			glMultiTexCoord1f(2, i->InitTimeStep);
	#endif
		glVertex2f((float)(i->Index % Width),
			(float)(i->Index / Height) + offsetY /* + offsetY shouldn't be */);
	}
	glEnd();

	cgGLDisableProfile(CG_PROFILE_FP30);

	VelocityPBuffers[VelocityCurrent]->EndCapture();
#endif

	// Render current position into texture.
	PositionPBuffers[PositionCurrent]->BeginCapture();

#if COMPLEX_INTITIALIZATION
	cgGLBindProgram(MovementInitFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	VelocityPBuffers[VelocityCurrent]->Bind();
#else
	cgGLBindProgram(InitValueFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);
#endif

	glBegin(GL_POINTS);
	for (i = NewParticles.begin(); i != NewParticles.end(); i++)
	{
		glMultiTexCoord3fv(0, i->Position.v);
#if COMPLEX_INTITIALIZATION
		glMultiTexCoord2f(1, (float)(i->Index % Width), (float)(i->Index / Height));
		glMultiTexCoord1f(2, i->InitTimeStep);
#endif
		glVertex2f((float)(i->Index % Width),
			(float)(i->Index / Height) + offsetY /* + offsetY shouldn't be */);
	}
	glEnd();

	cgGLDisableProfile(CG_PROFILE_FP30);

	PositionPBuffers[PositionCurrent]->EndCapture();

	// Render time of death into texture.
	DeathTimePBuffer->BeginCapture();

	cgGLBindProgram(InitValueFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glBegin(GL_POINTS);
	for (i = NewParticles.begin(); i != NewParticles.end(); i++)
	{
		glTexCoord1f(i->TimeOfDeath);
		glVertex2f((float)(i->Index % Width),
			(float)(i->Index / Height) + offsetY /* + offsetY shouldn't be */);
	}
	glEnd();

	cgGLDisableProfile(CG_PROFILE_FP30);

	DeathTimePBuffer->EndCapture();


	// Free array.
	NewParticles.clear();
}

void ParticleSystem::CleanUpParticles()
{
	while (ActiveParticles.top().TimeOfDeath < TimeNow)
	{
		int i = ActiveParticles.top().Index;
		ActiveParticles.pop();

		AvailableIndices.push(i);

		UsedIndices[i] = false;

		if (i >= MaxUsedIndex)
			while (!UsedIndices[MaxUsedIndex])
				MaxUsedIndex--;
	}
}

void ParticleSystem::DisplayCollisionSphere()
{
	glPushAttrib(GL_ENABLE_BIT | GL_LIGHTING_BIT);

	glPushMatrix();
	glTranslatef(SpherePosition[0], SpherePosition[1], SpherePosition[2]);

	float4 lightDirection[2];
	lightDirection[0] = -float4(1.0f, -2.0f, -1.0f, 0.0f);
	lightDirection[1] = -float4(-1.0f, 0.0f, -1.0f, 0.0f);
	float4 ambient(0.04f, 0.09f, 0.36f, 1.0f);
	float4 diffuse(0.08f, 0.18f, 0.72f, 0.0f);
	float4 specular(0.50f, 0.50f, 0.60f, 0.0f); 
	float shininess = 50.0f;

	glEnable(GL_NORMALIZE);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, lightDirection[0].v);
	glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_POSITION, lightDirection[1].v);

	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient.v);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse.v);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular.v);
	glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);


	glutSolidSphere(1, 80, 40);


	glDisable(GL_LIGHTING);
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHT1);

	glPopMatrix();

	glPopAttrib();
}

void ParticleSystem::DisplayTerrain()
{
	glPushAttrib(GL_ENABLE_BIT | GL_LIGHTING_BIT);

	glPushMatrix();
	glScalef(4.0f / TerrainHeightField.get_width(), 4 * 1.0f / 255.0f, 4.0f / TerrainHeightField.get_height());
	glTranslatef(-TerrainHeightField.get_width() / 2.0f, 0.0, -TerrainHeightField.get_height() / 2.0f);

	float4 lightDirection[2];
	lightDirection[0] = float4(1.0f, -2.0f, -1.0f, 0.0f);
	lightDirection[1] = float4(-1.0f, 0.0f, -1.0f, 0.0f);
	float4 ambient(0.19f, 0.14f, 0.11f, 1.0f);
	float4 diffuse(0.38f, 0.28f, 0.22f, 0.0f);
	float4 specular(0.05f, 0.03f, 0.02f, 0.0f); 
	float shininess = 10.0f;

	glEnable(GL_NORMALIZE);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, lightDirection[0].v);
	glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_POSITION, lightDirection[1].v);

	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient.v);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse.v);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular.v);
	glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);


	TerrainSimpleMesh.Render();


	glDisable(GL_LIGHTING);
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHT1);

	glPopMatrix();

	glPopAttrib();
}

void ParticleSystem::DisplayFlowField()
{
	glLineWidth(1.0);

	glPushMatrix();
	glScalef(FlowFieldScale[0] / FlowFieldSize[0], FlowFieldScale[1] / 255.0f, FlowFieldScale[2] / FlowFieldSize[1]);
	glTranslatef(-FlowFieldSize[0] / 2.0f, 0.0, -FlowFieldSize[1] / 2.0f);

	cgGLSetParameter4f(Passthru_Scale_Param, 1, 1, 1, 1);
	cgGLSetParameter4f(Passthru_Bias_Param, 0, 0, 0, 0);

	cgGLBindProgram(PassthruFProg);
	cgGLEnableProfile(CG_PROFILE_FP30);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, FlowFieldTexture);

	glBegin(GL_QUADS);

	glTexCoord2i(0, 0);
	glVertex3i(0, 0, 0);
	glTexCoord2i(FlowFieldSize[0] - 1, 0);
	glVertex3i(FlowFieldSize[0] - 1, 0, 0);
	glTexCoord2i(FlowFieldSize[0] - 1, FlowFieldSize[1] - 1);
	glVertex3i(FlowFieldSize[0] - 1, 0, FlowFieldSize[1] - 1);
	glTexCoord2i(0, FlowFieldSize[1] - 1);
	glVertex3i(0, 0, FlowFieldSize[1] - 1);

	glEnd();

	cgGLDisableProfile(CG_PROFILE_FP30);

	glPopMatrix();
}

// Return logarithm to base 2. Works only for power of two values.
static int logd(int value)
{
	int l;
	for (l = 0; value >> l; l++);
	return l - 1;
}

void ParticleSystem::MergeSort()
{
	SortCurrentPass = 0;

	int logdSize = logd(GetSize());
	SortMaxPasses = (logdSize + 1) * logdSize / 2;

	SortPassBegin = SortPassEnd;
	SortPassEnd = (SortPassBegin + SortPassesPerFrame) % SortMaxPasses;

	/*SortPBuffers[SortCurrent]->BeginCapture();
	glClear(GL_COLOR_BUFFER_BIT);
	DrawImage();
	SortPBuffers[SortCurrent]->EndCapture();*/

	cgGLSetParameter2f(SortRecursion_Size_Param, (float)Width, (float)Height);
	cgGLSetParameter2f(SortEnd_Size_Param, (float)Width, (float)Height);

	DoMergeSortPass(GetSize());

	/*
	SortPBuffers[SortCurrent]->BeginCapture();
	glReadPixels(0, 0, Width, Height, GL_RED, GL_FLOAT, (float*)DebugBuffer);
	SortPBuffers[SortCurrent]->EndCapture();

	for (int i = 0; i < GetSize(); i++)
		//printf("%i, ", (int)(DebugBuffer[i] * 255 + 0.5));
		printf("%.2f, ", DebugBuffer[i]);
	printf("\n");


	for (int i = 0; i < GetSize() - 1; i++)
		if (DebugBuffer[i] < DebugBuffer[i + 1])
			//printf("Error at %i: %i < %i\n", i, (int)(DebugBuffer[i] * 255 + 0.5), (int)(DebugBuffer[i + 1] * 255 + 0.5));
			printf("Error at %i: %.2f < %.2f\n", i, DebugBuffer[i], DebugBuffer[i + 1]);
	*/
}

void ParticleSystem::DoMergeSortPass(int count)
{
	//printf("mergeSort: count=%i\n", count);

	if (count > 1)
	{
		DoMergeSortPass(count / 2);
		DoMergePass(count, 1);
	}

	//printf("mergeSort: end\n");
}

void ParticleSystem::DoMergePass(int count, int step)
{
	if (count > 2)
	{
		DoMergePass(count / 2, step * 2);

		SortCurrentPass++;

		if (SortPassBegin < SortPassEnd)
		{
			if (SortCurrentPass < SortPassBegin || SortCurrentPass >= SortPassEnd)
				return;
		}
		else
		{
			if (SortCurrentPass < SortPassBegin && SortCurrentPass >= SortPassEnd)
				return;
		}

		//printf("%02i: mergeRec: count=%i, step=%i\n", SortCurrentPass, count, step);

		SortPBuffers[SortDest]->BeginCapture();

		cgGLBindProgram(SortRecursionFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		SortPBuffers[SortCurrent]->Bind();

		cgGLSetParameter1f(SortRecursion_Step_Param, (float)step);
		cgGLSetParameter1f(SortRecursion_Count_Param, (float)count);

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		SortPBuffers[SortDest]->EndCapture();

		int temp = SortCurrent;
		SortCurrent = SortDest;
		SortDest = temp;
	}
	else
	{
		SortCurrentPass++;

		if (SortPassBegin < SortPassEnd)
		{
			if (SortCurrentPass < SortPassBegin || SortCurrentPass >= SortPassEnd)
				return;
		}
		else
		{
			if (SortCurrentPass < SortPassBegin && SortCurrentPass >= SortPassEnd)
				return;
		}

		//printf("%02i: mergeEnd: count=%i, step=%i\n", SortCurrentPass, count, step);

		SortPBuffers[SortDest]->BeginCapture();

		cgGLBindProgram(SortEndFProg);
		cgGLEnableProfile(CG_PROFILE_FP30);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		SortPBuffers[SortCurrent]->Bind();

		cgGLSetParameter1f(SortEnd_Step_Param, (float)step);

		DrawQuad();

		cgGLDisableProfile(CG_PROFILE_FP30);

		SortPBuffers[SortDest]->EndCapture();

		int temp = SortCurrent;
		SortCurrent = SortDest;
		SortDest = temp;
	}
}
