// Copyright (c) 2004-2005 Lutz Latta
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#ifndef ParticleSystem_H
#define ParticleSystem_H

#include <vector>
#include <queue>
#include <stack>

#include <Cg/cgGL.h>
#include <glh/glh_array.h>
#include <shared/RenderTexture.h>
#include <shared/data_path.h>
#include "VectorMath.h"
#include "RenderVertexArray.h"
#include "RenderToVA.h"
#include "SimpleMesh.h"

using namespace std;
using namespace glh;

inline float randomRange(float rangeMin, float rangeMax)
{
	return ((float)rand() / RAND_MAX) * (rangeMax - rangeMin) + rangeMin ;
}

template <typename Type>
class stack_vector : public stack<Type, vector<Type> >
{
public:
	void reserve(typename stack_vector<Type>::size_type count)
	{
		c.reserve(count);
	}
};

template <typename Type, typename Compare=less<Type> >
class priority_queue_vector : public priority_queue<Type, vector<Type>, Compare>
{
public:
	void reserve(typename priority_queue_vector<Type>::size_type count)
	{
		c.reserve(count);
	}
};

struct ActiveParticle
{
	float TimeOfDeath;
	int Index;
};

namespace std
{
	template <>
	struct greater<ActiveParticle>
	{
		bool operator() (const ActiveParticle& left, const ActiveParticle& right)
		{
			return left.TimeOfDeath > right.TimeOfDeath;
		}
	};
}

struct NewParticle
{
	int Index;
	float3 Position;
	float3 Velocity;
	float TimeOfDeath;
	float InitTimeStep;
};

struct ParticleVertexData
{
	float TimeOfBirth;
	float LifeTime;
	float AttributeType;
	float Index;
};

struct ParticleAttributesType
{
	float2 PointSizeRange;
	float4 OrientationStartRange_VelocityRange;
	float4 TileScale_TileOffset;
	float4 ColorAnimationSegments;
	float4 ColorAnimationFunctions[6];
};

struct ParticleAttributesParametersType
{
	CGparameter PointSizeRange;
	CGparameter OrientationStartRange_VelocityRange;
	CGparameter TileScale_TileOffset;
	CGparameter ColorAnimationSegments;
	CGparameter ColorAnimationFunctions;
};

struct ParticleEmitterData
{
	float3 Position;
	float3 PreviousPosition;

	float3 PositionMaxOffset;

	float2 InitialVelocityMinMax;
	float3 InitialDirectionCenter;
	float InitialDirectionMaxAngle;

	float2 LifeTimeMinMax;

	int AttributeType;

	float2 NumOfNewParticlesMinMax;
};

class ParticleSystem
{

public:

	enum BlendFunctionType
	{
		BF_ADD,
		BF_ALPHA,
		BF_PREMULTIPLIED_ALPHA,

		BF_NUMOF
	};


	ParticleSystem();
	~ParticleSystem();

	void LoadConfigSetup(const string& setupPrefix, bool allowResize = true);
	void UnloadConfigSetup(bool allowResize = true);

	int GetWidth() { return Width; }
	int GetHeight() { return Height; }
	int GetSize() { return Width * Height; }
	int GetNumOfActiveParticles() { return ActiveParticles.size(); }

	int GetNumOfEmitters() { return Emitters.size(); }
	int GetNumOfAttributeTypes() { return ParticleAttributes.size(); }

	void SetViewerPosition(const float3& position);
	const float3& GetViewerPosition();
	void SetViewerRotation(const float2& rotation);
	const float2& GetViewerRotation();

	void SetSpherePosition(const float3& position);
	const float3& GetSpherePosition();

	void SetEmitterPosition(int emitterIndex, const float3& position);
	const float3& GetEmitterPosition(int emitterIndex);

	void SetEmitterDirection(int emitterIndex, const float3& direction);
	const float3& GetEmitterDirection(int emitterIndex);

	void SetLocalForcePosition(int localForceIndex, const float3& position);
	const float3& GetLocalForcePosition(int localForceIndex);
	int GetNumOfLocalForces() { return (int)LocalForcePositions.size(); }

	const float4& GetClearColor() { return ClearColor; }

	void SetPerspective(float fieldOfViewY, int screenHeight);

	void SetMenuKeys(const bool menuKeys[256]) { memcpy(MenuKeys, menuKeys, sizeof(MenuKeys)); }
	const bool* GetMenuKeys() { return MenuKeys; }


	// Initialize particles.
	void Reset();

	// Step the simulation.
	void TimeStep(float dT);

	void ReadBack();

	// display points using render-to-vertex array
	void DisplayPoints();

	// display position and velocity textures
	void DisplayTextures();

	void DisplayCollisionSphere();
	void DisplayTerrain();

	void DisplayFlowField();

	void MergeSort();

private:

	RenderTexture* CreateRenderTexture(int numOfBits, int numOfChannels);
	CGprogram loadProgram(CGcontext context, CGprofile profile, char *filename, data_path &path, char* entryPoint = NULL);

	void DrawImage();
	void DrawQuad();

	// display float texture using texturing
	void DisplayTexture(RenderTexture *renderTexture);

	void AllocateParticle(const float3& position, const float3& velocity, float lifeTime, int attributeType, float initTimeStep);

	void InitNewParticles();
	void CleanUpParticles();

	void DoMergeSortPass(int count);
	void DoMergePass(int count, int step);

	int Width;
	int Height;

	float TimeNow;

	float3* PositionBuffer;
	float* DebugBuffer;

	RenderableVertexArray* VertexArray[1];
	int CurrentVertexArray;

	RenderVertexArray* RenderVertexBuffer;
	ParticleVertexData* ParticleVertices;
    GLuint ParticleVertexBuffer;

	// Allocation data.
	//stack_vector<int> AvailableIndices;
	priority_queue_vector<int, greater<int> > AvailableIndices;
	priority_queue_vector<ActiveParticle, greater<ActiveParticle> > ActiveParticles;

	vector<bool> UsedIndices;
	int MaxUsedIndex;

	vector<NewParticle> NewParticles;

	vector<ParticleEmitterData> Emitters;

	// float pbuffers representing current and previous positions, plus destination
	RenderTexture* PositionPBuffers[3];
	int PositionCurrent;
	int PositionPrevious;
	int PositionDest;

	RenderTexture* VelocityPBuffers[2];
	GLuint VelocityTextures[2];
	int VelocityCurrent;
	int VelocityDest;

	RenderTexture* DeathTimePBuffer;
	GLuint DeathTimeTexture;

	RenderTexture* SortPBuffers[2];
	GLuint SortTextures[2];
	int SortCurrent;
	int SortDest;

	int SortCurrentPass;
	int SortPassBegin;
	int SortPassEnd;
	int SortPassesPerFrame;
	int SortMaxPasses;

	CGprogram MovementVerletFProg;
	CGparameter MovementVerlet_TimeStep_Param;
	CGparameter MovementVerlet_SphereCenter_Param;

	CGprogram MovementFProg;
	CGparameter Movement_TimeStep_Param;

	CGprogram MovementInitFProg;

	CGprogram VelocityFProg;
	CGparameter Velocity_TimeStep_Param;
	CGparameter Velocity_SphereCenter_Param;
	CGparameter Velocity_LocalAttractorPosition_Param;

	CGprogram VelocityInitFProg;
	CGparameter VelocityInit_SphereCenter_Param;
	CGparameter VelocityInit_LocalAttractorPosition_Param;

	CGprogram PassthruFProg;
	CGparameter Passthru_Scale_Param;
	CGparameter Passthru_Bias_Param;

	CGprogram InitValueFProg;

	CGprogram InvalidateDeadFProg;
	CGparameter InvalidateDead_CurrentTime_Infinity_Param;

	CGprogram SortRecursionFProg;
	CGprogram SortEndFProg;
	CGparameter SortRecursion_Size_Param;
	CGparameter SortRecursion_Step_Param;
	CGparameter SortRecursion_Count_Param;
	CGparameter SortEnd_Size_Param;
	CGparameter SortEnd_Step_Param;

	CGprogram DistanceSortInitFProg;
	CGprogram DistanceSortComputeFProg;
	CGprogram DistanceSortLookupFProg;
	CGparameter DistanceSortCompute_ViewerPosition_Param;

	CGprogram DisplayFProg;
	CGprogram DisplayVProg; 
	CGparameter DisplayVP_ModelViewProj_Param;
	CGparameter DisplayVP_ModelView_Param;
	CGparameter DisplayVP_TanHalfFOVTimesScreenSize_Param;
	CGparameter DisplayVP_CurrentTime_Param;
	CGparameter DisplayVP_RandomTable_Param;

	float LastInvalidateDeadTime;

	vector<ParticleAttributesType> ParticleAttributes;
	vector<ParticleAttributesParametersType> ParticleAttributesParameters;

	vector<float4> RandomTable;

	float FieldOfViewY;
	int ScreenHeight;
	float3 ViewerPosition;
	float2 ViewerRotation;
	float4 ClearColor;

	BlendFunctionType BlendFunction;

	bool DoSorting;

	string DiffuseTextureName;
	GLuint DiffuseTexture;

	float3 SpherePosition;

	GLuint TerrainTexture;
	array2<unsigned char> TerrainHeightField;
	SimpleMesh TerrainSimpleMesh;

	string FlowFieldTextureName;
	GLuint FlowFieldTexture;
	int2 FlowFieldSize;
	float3 FlowFieldScale;

	vector<float3> LocalForcePositions;

	bool MenuKeys[256];
};

#endif
